﻿namespace DLC_Tool
{
    using System;
    using System.Collections.Generic;
    using System.IO;
    using System.Text;
    using System.Windows.Forms;

    internal static class Program
    {
        public static readonly Dictionary<byte, string> CharNames = new Dictionary<byte, string>()
        {
            { 0, "ZACK" }, { 1, "TINA" }, { 2, "JANNLEE" }, { 3, "EIN" }, { 4, "HAYABUSA" }, { 5, "KASUMI" }, { 6, "GENFU" }, { 7, "HELENA" }, { 8, "LEON" }, { 9, "BASS" },
            { 10, "KOKORO" }, { 11, "HAYATE" }, { 12, "LEIFANG" }, { 13, "AYANE" }, { 14, "ELIOT" }, { 15, "LISA" }, { 16, "ALPHA152" }, { 19, "BRAD" }, { 20, "CHRISTIE" },
            { 21, "HITOMI" }, { 24, "BAYMAN" }, { 29, "RIG" }, { 30, "MILA" }, { 31, "AKIRA" }, { 32, "SARAH" }, { 33, "PAI" }, { 39, "MOMIJI" }, { 40, "RACHEL" },
            { 41, "JACKY" }, { 42, "MARIE" }, { 43, "PHASE4" }, { 44, "NYOTENGU" }, { 45, "HONOKA" }, { 46, "RAIDOU" }
        };

        public static readonly List<byte> FemaleIDs = new List<byte> { 1, 5, 7, 10, 12, 13, 15, 16, 20, 21, 30, 32, 33, 39, 40, 42, 43, 44, 45 };
        private static List<string> decNames, encNames;

        public static DLCData OpenBCM(string fileName)
        {
            var dlcData = new DLCData();
            using (var ms = new MemoryStream())
            {
                var br = new BinaryReader(ms);
                using (var fs = new FileStream(fileName, FileMode.Open, FileAccess.Read, FileShare.Read))
                {
                    fs.CopyTo(br.BaseStream);
                }

                br.BaseStream.Position = 0;
                dlcData.BcmVer = br.ReadByte();
                br.BaseStream.Position += 1;
                byte entriesCount = br.ReadByte();
                br.BaseStream.Position += 5;
                for (int i = 0; i < entriesCount; i++)
                {
                    var charEntry = new Character();
                    charEntry.ID = br.ReadByte();
                    charEntry.CostumeSlot = br.ReadByte();
                    charEntry.AddTexsCount = br.ReadByte();
                    byte hStylesCount = br.ReadByte();
                    for (int j = 0; j < hStylesCount; j++)
                    {
                        charEntry.HStyles.Add(new Hairstyle(br.ReadByte(), br.ReadByte()));
                    }

                    dlcData.Chars.Add(charEntry);
                }
            }

            return dlcData;
        }

        public static void SaveBCM(DLCData dlcData, string savePath, string dlcName)
        {
            ulong dlcNum;
            if (!ulong.TryParse(dlcName, out dlcNum))
            {
                throw new Exception("File name must contain only numeric characters!");
            }

            ulong checkSum = 0;
            for (int i = 0; i < dlcData.Chars.Count; i++)
            {
                checkSum += (ushort)(dlcData.Chars[i].ID * dlcData.Chars[i].CostumeSlot);
            }

            checkSum = ((dlcNum + 1) * (checkSum & 4095)) % ((dlcData.BcmVer * dlcNum) + 17);

            using (var ms = new MemoryStream())
            {
                var bw = new BinaryWriter(ms);
                bw.Write(dlcData.BcmVer);
                bw.BaseStream.Position += 1;
                bw.Write((byte)dlcData.Chars.Count);
                bw.BaseStream.Position += 1;
                bw.Write(checked((uint)checkSum));
                for (int i = 0; i < dlcData.Chars.Count; i++)
                {
                    bw.Write(dlcData.Chars[i].ID);
                    bw.Write(dlcData.Chars[i].CostumeSlot);
                    bw.Write(dlcData.Chars[i].AddTexsCount);
                    bw.Write((byte)dlcData.Chars[i].HStyles.Count);
                    for (int j = 0; j < dlcData.Chars[i].HStyles.Count; j++)
                    {
                        bw.Write(dlcData.Chars[i].HStyles[j].Hair);
                        bw.Write(dlcData.Chars[i].HStyles[j].Face);
                    }
                }

                bw.BaseStream.Position = 0;
                using (var fs = new FileStream(savePath, FileMode.Create, FileAccess.Write, FileShare.Read))
                {
                    bw.BaseStream.CopyTo(fs);
                }
            }
        }

        public static List<int> SaveBIN(DLCData dlcData, string savePath, string dlcName)
        {
            if (decNames == null)
            {
                using (var sr = new StreamReader(Path.GetDirectoryName(Application.ExecutablePath) + @"\dlc5lr.dat"))
                {
                    decNames = new List<string>();
                    encNames = new List<string>();
                    while (!sr.EndOfStream)
                    {
                        string[] inLine = sr.ReadLine().Split('\t');
                        encNames.Add(inLine[0]);
                        decNames.Add(inLine[1]);
                    }
                }
            }

            var nameIndexes = new List<int>();
            var fileNames = new List<string>();
            for (int i = 0; i < dlcData.Chars.Count; i++)
            {
                if ((dlcData.Chars[i].Files[0] == null) || (dlcData.Chars[i].Files[1] == null)
                    || (dlcData.Chars[i].Files[2] == null) || (dlcData.Chars[i].Files[11] == null))
                {
                    throw new Exception("Path to TMC/TMCL/---C/--P file for " + CharNames[dlcData.Chars[i].ID] + "[" + dlcData.Chars[i].CostumeSlot + "] not found!");
                }

                string decTmcName = CharNames[dlcData.Chars[i].ID] + "_DLCU_" + (dlcData.Chars[i].CostumeSlot + 1).ToString("D3") + ".TMC";
                int nameIndex = decNames.IndexOf(decTmcName);
                if (nameIndex == -1)
                {
                    throw new Exception("Filename for " + CharNames[dlcData.Chars[i].ID] + "[" + dlcData.Chars[i].CostumeSlot + "] not found in database!");
                }

                nameIndexes.Add(nameIndex);
                nameIndexes.Add(nameIndex + 1);
                nameIndexes.Add(nameIndex + 2);
                fileNames.Add(encNames[nameIndex]);
                fileNames.Add(encNames[nameIndex + 1]);
                fileNames.Add(encNames[nameIndex + 2]);
                if (dlcData.Chars[i].AddTexsCount > 1)
                {
                    for (int j = 0; j < dlcData.Chars[i].AddTexsCount; j++)
                    {
                        if ((dlcData.Chars[i].Files[3 + (j * 2)] == null) || (dlcData.Chars[i].Files[4 + (j * 2)] == null))
                        {
                            throw new Exception("Path to " + (j + 1).ToString("D3") + ".--H/--HL file for " + CharNames[dlcData.Chars[i].ID] + "["
                                + dlcData.Chars[i].CostumeSlot + "] not found!");
                        }

                        nameIndexes.Add(nameIndex + 3 + (j * 2));
                        nameIndexes.Add(nameIndex + 4 + (j * 2));
                        fileNames.Add(encNames[nameIndex + 3 + (j * 2)]);
                        fileNames.Add(encNames[nameIndex + 4 + (j * 2)]);
                    }
                }

                nameIndex = decNames.IndexOf(Path.ChangeExtension(decTmcName, ".--P"), nameIndex + 3);
                nameIndexes.Add(nameIndex);
                fileNames.Add(encNames[nameIndex]);
            }

            int headerSize = 40 + (fileNames.Count * 12);
            using (var ms = new MemoryStream())
            {
                var bw = new BinaryWriter(ms, new ASCIIEncoding());
                bw.BaseStream.Position = headerSize;
                bw.Write(("_output/costume_pack_" + dlcName + char.MinValue + "order" + char.MinValue).ToCharArray());
                var namesOffs = new List<int>();
                foreach (string fileName in fileNames)
                {
                    namesOffs.Add((int)bw.BaseStream.Position);
                    bw.Write((@"/" + fileName + char.MinValue).ToCharArray());
                }

                bw.BaseStream.Position = 0;
                bw.Write(0x4F4D464C);
                bw.Write(1);
                bw.Write(fileNames.Count);
                bw.Write(0x20);
                bw.Write(0x28);
                bw.Write(headerSize);
                bw.Write(namesOffs[0]);
                bw.Write(0);
                bw.Write(0);
                bw.Write(headerSize);
                for (int i = 0; i < namesOffs.Count; i++)
                {
                    bw.Write(0);
                    bw.Write(i);
                    bw.Write(namesOffs[i]);
                }

                bw.BaseStream.Position = 0;
                using (var fs = new FileStream(savePath, FileMode.Create, FileAccess.Write, FileShare.Read))
                {
                    bw.BaseStream.CopyTo(fs);
                }
            }

            return nameIndexes;
        }

        public static List<uint> SaveLNK(DLCData dlcData, string savePath, int fileCount)
        {
            int headerSize = 32 + (fileCount * 32);
            if (headerSize % 2048 > 0)
            {
                headerSize = headerSize - (headerSize % 2048) + 2048;
            }

            var fileSizes = new List<uint>();
            var fileOffs = new List<uint>();
            using (var fs = new FileStream(savePath, FileMode.Create, FileAccess.Write, FileShare.Read))
            {
                fs.Position = headerSize;
                for (int i = 0; i < dlcData.Chars.Count; i++)
                {
                    int orderLimit = 2;
                    if (dlcData.Chars[i].AddTexsCount > 1)
                    {
                        orderLimit += dlcData.Chars[i].AddTexsCount * 2;
                    }

                    for (int j = 0; j < 12; j++)
                    {
                        if ((j > orderLimit) && (j < 11))
                        {
                            continue;
                        }

                        fileOffs.Add((uint)fs.Position);
                        using (var fs2 = new FileStream(dlcData.Chars[i].Files[j], FileMode.Open, FileAccess.Read, FileShare.Read))
                        {
                            fileSizes.Add((uint)fs2.Length);
                            fs2.CopyTo(fs);
                        }

                        if (fs.Position % 2048 > 0)
                        {
                            fs.Position = fs.Position - (fs.Position % 2048) + 2048;
                        }
                    }
                }

                var bw = new BinaryWriter(fs);
                if (fs.Length % 2048 > 0)
                {
                    bw.BaseStream.Position -= 1;
                    bw.Write((byte)0);
                }

                bw.BaseStream.Position = 0;
                bw.Write(0x4D444350);
                bw.Write(0);
                bw.Write(fileCount);
                bw.Write(0);
                bw.Write((uint)bw.BaseStream.Length);
                bw.Write(0);
                bw.Write(2048);
                bw.Write(0);
                for (int i = 0; i < fileCount; i++)
                {
                    bw.Write(fileOffs[i]);
                    bw.Write(0);
                    bw.Write(fileSizes[i]);
                    bw.Write(0);
                    bw.Write(fileSizes[i]);
                    bw.Write(0);
                    bw.Write(0);
                    bw.Write(0);
                }
            }

            return fileSizes;
        }

        public static void SaveBLP(string savePath, List<int> nameIndexes, List<uint> fileSizes)
        {
            using (var ms = new MemoryStream())
            {
                ms.Position = 16;
                var bw = new BinaryWriter(ms);
                for (int i = 0; i < nameIndexes.Count; i++)
                {
                    bw.Write(nameIndexes[i]);
                    bw.Write(fileSizes[i]);
                    bw.Write(0);
                    bw.Write(0);
                    bw.Write(0);
                }

                bw.BaseStream.Position = 0;
                bw.Write(0x46495031);
                bw.Write(nameIndexes.Count);
                bw.Write(16);
                bw.Write((int)bw.BaseStream.Length);
                bw.BaseStream.Position = 0;
                using (var fs = new FileStream(savePath, FileMode.Create, FileAccess.Write, FileShare.Read))
                {
                    bw.BaseStream.CopyTo(fs);
                }
            }
        }

        [STAThread]
        private static void Main()
        {
            Application.EnableVisualStyles();
            Application.SetCompatibleTextRenderingDefault(false);
            Application.Run(new MainForm());
        }
    }
}